/**
 * This file provides class DerivativeContainer.
 * This class represents an enclosure as a union (list) of doubletons.
 * Each doubleton represent matrix in the form
 * D + C*R0 + B*R
 * where D,C is a point matrix
 * B - is an orthogonal matrix
 * R0, R  are interval matrices centered at zero.
 * 
 * The main method 'affineTransformation' computes the value of
 * M*(D + C*R0 + B*R)
 * assuming that M is almost point matrix.
 * It is better to evaluate this product in the following order
 * M*D + (M*C)*R0 + (M*B)*R.
*/

#ifndef DW_DER_CONTAINER
#define DW_DER_CONTAINER

template<class C1SetType>
class DerContainer : public std::list<C1SetType>
{
public:
  typedef typename C1SetType::MatrixType MatrixType;
  typedef typename std::list<C1SetType>::iterator iterator;


  // for a given iterator b we evaluate M*D + (M*C)*R0 + (M*B)*R.
  static MatrixType affineTransformation(iterator b, const MatrixType& M)
  {
    return M*b->get_D() + (M*b->get_Cjac())*b->get_R0() + (M*b->get_Bjac())*b->get_R();
  }

  // evaluate  M*D + (M*C)*R0 + (M*B)*R for each element 
  // from the list and take the inetrval hull.
  MatrixType affineTransformation(const MatrixType& M)
  {
    iterator b = this->begin(), e = this->end();
    MatrixType result = affineTransformation(b,M);
    ++b;
    while(b!=e)
    {
      result = intervalHull(result,affineTransformation(b,M));
      ++b;
    }
    return result;
  }

  // as before. But instead of taking interval hull we return 
  // list of matrices  M*D + (M*C)*R0 + (M*B)*R
  void affineTransformation(const MatrixType& M, std::list<MatrixType>& result)
  {
    result.clear();
    iterator b = this->begin(), e = this->end();
    while(b!=e)
    {
      result.push_back(affineTransformation(b,M));
      ++b;
    }
  }
};

#endif
