#ifndef DW_SMALE_WILLIAMS_MAP_H
#define DW_SMALE_WILLIAMS_MAP_H

#include <list>

// include all necessary solvers and data structures
// from the CAPD library

#include "capd/vectalg/Vector.hpp"
#include "capd/vectalg/Matrix.hpp"
#include "capd/vectalg/Norm.hpp"
#include "capd/map/Map.hpp"
#include "capd/map/CnMap.hpp"
#include "capd/dynsys/CnTaylor.hpp"
#include "capd/dynsys/TaylorHOE.hpp"
#include "capd/dynset/TimeSet.h"
#include "capd/dynset/Pped2Set.hpp"
#include "capd/dynset/Intv2Set.hpp"
#include "capd/dynset/Rect2Set.hpp"
#include "capd/dynset/C1Rect2.hpp"
#include "capd/dynset/C11Rect2.hpp"
#include "capd/poincare/TimeMap.hpp"

// this is for optimization of computing of derivatives
#include "DerContainer.h"

const int SW_DIM = 4;
typedef long double Real;

// define types for vectors, matrixes, solvers etc.
typedef capd::vectalg::Vector<interval,SW_DIM> SWVector;
typedef capd::vectalg::Matrix<interval,SW_DIM,SW_DIM> SWMatrix;
typedef capd::vectalg::MaxLNorm<SWVector,SWMatrix> ELNorm;

typedef capd::vectalg::Vector<Real,SW_DIM> LDSWVector;
typedef capd::vectalg::Matrix<Real,SW_DIM,SW_DIM> LDSWMatrix;

typedef capd::map::CnMap<SWMatrix> SWMap;
typedef capd::dynsys::TaylorHOE<SWMap> Solver;
typedef capd::dynsys::NonAutDynSys<Solver> NASolver;

typedef capd::map::CnMap<LDSWMatrix> LDSWMap;
typedef capd::dynsys::BasicCnTaylor<LDSWMap> LDSolver;

typedef capd::dynset::Rect2Set<SWMatrix> SWRect2Set;
typedef capd::dynset::C1Rect2<SWMatrix> SWC1Rect2;
typedef capd::dynset::TimeSet<SWRect2Set> TimeRect2;
typedef capd::dynset::TimeSet<SWC1Rect2> TimeC1Rect2;

typedef capd::poincare::TimeMap<NASolver> TimeMap;

// this is vector field for the Kuznetsov system
static std::string vf = "\
par:twoPi,omega,twoOmega,A,eps,epsByOmega,epsByTwoOmega,twoPiByT;\
var:x,u,y,v,t;\
fun:\
omega*u,\
-omega*x+(A*cos(twoPiByT*t)-x^2)*u+epsByOmega*y*cos(omega*t),\
twoOmega*v,\
epsByTwoOmega*x^2-twoOmega*y-(A*cos(twoPiByT*t)+y^2)*v;\
";


// ----------------------------- SmaleWilliamsMap ---------------------
/**
 * This class represents Poincare map for the Kuznetsov system.
 * It provides required methods for computation of Poincare map
 * and its derivative.
*/
class SmaleWilliamsMap
{
public:
  typedef SWVector VectorType;
  typedef SWMatrix MatrixType;
  typedef DerContainer<TimeC1Rect2> DerivativeContainer;

  // This is contructor of the class.
  // Here we declare vector fields, solvers and Poincare map.
  SmaleWilliamsMap(interval omega, double _period, double A, double eps, int order=15)
    :   factor(1), c1factor(1), period(_period)
  {
    // for each thread on the machine we define own vector field object.
    // both rigorous and nonrigorous
    vectorField = new SWMap* [omp_get_max_threads()];
    ldVectorField = new LDSWMap* [omp_get_max_threads()];

    // for each thread on the machine we define own solver object.
    // both rigorous and nonrigorous
    T = new Solver* [omp_get_max_threads()];
    ldT = new LDSolver* [omp_get_max_threads()];

    // for each thread on the machine we define own Poincare map object.
    // both rigorous and nonrigorous
    naT = new NASolver* [omp_get_max_threads()];
    P = new TimeMap* [omp_get_max_threads()];

    // Arrays for objects are allocated. Define objects in arrays.
    for (int i=0; i<omp_get_max_threads();++i)
    {
      // define vector fields
      vectorField[i] = new SWMap(vf);
      ldVectorField[i] = new LDSWMap(vf);

      // set parameters of vector fields (rigorous)
      vectorField[i]->setParameter("twoPi",2*interval::pi());
      vectorField[i]->setParameter("omega",omega);
      vectorField[i]->setParameter("twoOmega",2*omega);
      vectorField[i]->setParameter("A",A);
      vectorField[i]->setParameter("eps",eps);
      vectorField[i]->setParameter("epsByOmega",eps/omega);
      vectorField[i]->setParameter("epsByTwoOmega",0.5*eps/omega);
      vectorField[i]->setParameter("twoPiByT",2*interval::pi()/period);

      // set parameters of vector fields (long double version)
      ldVectorField[i]->setParameter("twoPi",2*M_PI);
      ldVectorField[i]->setParameter("omega",omega.mid().leftBound());
      ldVectorField[i]->setParameter("twoOmega",2*omega.mid().leftBound());
      ldVectorField[i]->setParameter("A",A);
      ldVectorField[i]->setParameter("eps",eps);
      ldVectorField[i]->setParameter("epsByOmega",eps/omega.mid().leftBound());
      ldVectorField[i]->setParameter("epsByTwoOmega",0.5*eps/omega.mid().leftBound());
      ldVectorField[i]->setParameter("twoPiByT",2*M_PI/period);

      // define solvers with a given vector field, order of Taylor method and StepControl
      T[i] = new Solver(*(vectorField[i]),order,1e-3,capd::dynsys::ILastTermsStepControl(1,1e-10));
      naT[i] = new NASolver(*(T[i]));
      P[i] = new TimeMap(*(naT[i]));

      ldT[i] = new LDSolver(*(ldVectorField[i]),order,2e-2);
    }

    // these are constants from the article that define linear change of coordinates for the attractor
    interval cux = 0.438;
    interval cyx = -0.042;
    interval cyu = 0.226;
    interval cvx = -0.218;
    interval cvu = 0.029;
    interval cvy = -0.118;
    interval r0 = 0.812;
    interval r1 = 0.721;

    // Here we set the matrix of linear change of coordinates.
    // It is triangular
    coordChange(1,1) = 1./r0;
    coordChange(2,1) = -cux/r1;
    coordChange(2,2) = 1./r1;
    coordChange(3,1) = -cyx;
    coordChange(3,2) = -cyu;
    coordChange(3,3) = 1;
    coordChange(4,1) = -cvx;
    coordChange(4,2) = -cvu;
    coordChange(4,3) = -cvy;
    coordChange(4,4) = 1;

    // Rigorous inverse of the above matrix.
    invCoordChange = capd::matrixAlgorithms::invLowerTriangleMatrix(coordChange);

    // Copy of the above matrixes to point matrices - for nonrigorous computations 
    // used for example, when computing of coordinate systems or periodic points.
    for(int i=1;i<=4;++i)
      for(int j=1;j<=4;++j)
      {
        ldCoordChange(i,j) = coordChange(i,j).mid().leftBound();
        ldInvCoordChange(i,j) = invCoordChange(i,j).mid().leftBound();
      }
  }

  // This is a method that sets the order of the Taylor method to each solver for each thread.
  void setOrder(int newOrder)
  {
    for (int i=0; i<omp_get_max_threads();++i)
    {
      naT[i]->setOrder(newOrder);
      ldT[i]->setOrder(newOrder);
    }
  }

  // This is a function that sets a factor - this is a very technical parameter 
  // of doubleton sets that specifies when to rearange representation.
  void setFactor(double _factor)
  {
    factor = _factor;
  }

  // This function sets new parameters of step control method
  // for each solver.
  void setStepControl(capd::dynsys::ILastTermsStepControl sc)
  {
    for(int i=0;i<omp_get_max_threads();++i)
      T[i]->setStepControl(sc);
  }


  // This is a nonrigorous routine. 
  // It computes Poincare map P(u).
  LDSWVector operator()(LDSWVector u)
  {
    capd::rounding::DoubleRounding::roundNearest();
    // put the point to the main coordinate system
    u = ldInvCoordChange*u;
    int i = omp_get_thread_num();
    Real time = 0.;
    Real step =  ldT[i]->getStep();
    bool done=true;

    // integrate it for time 'period'
    while(done)
    {
      ldT[i]->differentiateTime();
      ldT[i]->setCurrentTime(time);
      if(time+step>period)
      {
        // here we make last smaller step to reach time 'period'
        ldT[i]->setStep(period-time);
        done = false;
      }
      // Here we compute the next point on the trajectory
      // using the solver ldT[i] - with long double precision.
      u = (*ldT[i])(u);
      time += step;
    }
    ldT[i]->setStep(step);

    // Put the result to the new coordinate system and return.
    return ldCoordChange*u;
  }

  // This is a nonrigorous routine. 
  // It computes simultanously Poincare map and its derivative.

  LDSWVector operator()(LDSWVector u, LDSWMatrix& result)
  {
    capd::rounding::DoubleRounding::roundNearest();

    // the initial condition for solving of derivative 
    // of Poincare map is usually Identity.
    // Since we have different coordinate system we can start from 
    // the matrix of coordinate system.
    result = ldInvCoordChange;

    // put the point to the main coordinate system
    u = ldInvCoordChange*u;

    int i = omp_get_thread_num();
    Real time = 0.;
    Real step =  ldT[i]->getStep();
    bool done=true;

    // integrate trajectory and monodromy matrix for the time 'period'
    while(done)
    {
      ldT[i]->differentiateTime();
      ldT[i]->setCurrentTime(time);
      if(time+step>period)
      {
        ldT[i]->setStep(period-time);
        done = false;
      }
      LDSWMatrix temp;
      // Here we compute the next point on the trajectory and the monodromy matrix
      // using the solver ldT[i] - with long double precision.
      // u is just the next point returned by the Solver from the CAPD
      u = (*ldT[i])(u,temp);
      // and we have to propagate derivative just by the composition, i.e.
      // newMonodromyMatrix = MonodromyMatrixForOneStep * previousMonodromyMatrix
      result = temp*result;
      time += step;
    }
    ldT[i]->setStep(step);

    // put both derivative and point to the required coordinate system
    result = ldCoordChange*result;
    return ldCoordChange*u;
  }


  // This is rigorous routine. For an interval vector u
  // it computes an interval vector that encloses P(u).
  VectorType operator()(const VectorType& u)
  {
    // To efficiently solve ODE we use doubleton representation of the sets, i.e. 
    // the sets are represented as
    // x + C*r0 + B*r
    // where
    // x - is a point vector
    // C - is a point matrix
    // B - is an orthogonal matrix
    // r0, r are interval vectors centered at zero.
    VectorType x=u, r0(4), r(4);
    
    // split means x = center(x), r0 = x-center(x)
    split(x,r0);

    // change the coordinates
    x = invCoordChange*x;
    // and split again
    split(x,r);

    // What we computed is
    // invCoordinateSystem * u 
    // that is covered by
    // x + invCoordinateSytem* r0 + Identity*r

    // define doubleton representation of the above
    SWRect2Set s(x,invCoordChange,r0,r,factor);
    TimeRect2 ts(s,0.);

    // and rigorously integrate it over the time 'period'
    // here we call the solver from the CAPD library
    (*(P[omp_get_thread_num()]))(period,ts);

    // finally, we put the result to the coordinates 'coordSystem'
    // this is computed by the member function of the class TimeSet
    // 'affineTransformation'
    return ts.affineTransformation(coordChange,SWVector());
  }


  // This is rigorous routine. It computes derivative of the Poincare map.
  // The result is added to the container dc.
  // The derivative is exporessed in the 'invCoordSystem*localCoord' in the domain.
  // In the range the coordinates are set to 'coordSystem'.
  void eval(const VectorType& u, const MatrixType& localCoord, DerivativeContainer& dc)
  {
    // Logarithmic norm, necessary to rigorously integrate variational equations.
    ELNorm N;

    // compute doubleton represenation - see function above.
    VectorType x=u, r0(4), r(4);
    split(x,r0);
    x = invCoordChange*x;
    split(x,r);
    SWC1Rect2 s(x,invCoordChange,r0,N,factor);
    s.set_r(r);

    // here we set doubleton representation for the derivative.
    s.set_D(invCoordChange*localCoord);
    TimeC1Rect2 ts(s,0.);

    // integrate trajetory and monodromy matrix for time 'period'
    (*(P[omp_get_thread_num()]))(period,ts);

    // put the obtained derivative to the coordinate system 'coordSystem'
    ts.set_D(coordChange* ts.get_D());
    ts.set_Cjac(coordChange * ts.get_Cjac());
    ts.set_Bjac(coordChange * ts.get_Bjac());

    ts.set_x(coordChange* ts.get_x());
    ts.set_C(coordChange * ts.get_C());
    ts.set_B(coordChange * ts.get_B());

    // and add to the container 
    dc.push_back(ts);
  }

  // This is rigorous routine. It computes derivative of Poincare map at u.
  // It just calls the above routine 'eval'
  // with localCoord=Identity
  // and extracts result from the container.
  MatrixType operator[](VectorType u)
  {
    DerivativeContainer dc;
    this->eval(u,SWMatrix::Identity(4),dc);
    std::list<MatrixType> derEnc;
    dc.affineTransformation(SWMatrix::Identity(4),derEnc);
    return *derEnc.begin();
  }

  // This is just for required interface. The last argument is ignored.
  // It might be used for subdivision, if necessary.
  void eval(VectorType u, const MatrixType& localCoord, DerivativeContainer& dc, int )
  {
    dc.clear();
    this->eval(u,localCoord,dc);
  }

  // This is just to satisfy interface. 
  // We rigorously compute P(u) and the result is a list of interval vectors.
  // This is in the case, subdivision is necessary.
  // Here the resulting list will always contain one element.
  void eval(VectorType u, std::list<VectorType>& result)
  {
    result.clear();
    u = (*this)(u);
    result.push_back(u);
  }


  // Members of the class
  double factor, c1factor;
  
  // coorinates of the attractor, rigorous and long double
  SWMatrix coordChange, invCoordChange;
  LDSWMatrix ldCoordChange, ldInvCoordChange;

  double period;

  // vector fields, solvers and Poicnare maps.
  SWMap** vectorField;
  LDSWMap** ldVectorField;
  LDSWMap** invLdVectorField;

  Solver** T;
  LDSolver** ldT;
  LDSolver** invLdT;

  NASolver** naT;

  TimeMap** P;
};


#endif
