/**
 * This file implements Algorithm 2 from the article.
 * It computes computes coordinate system in each vertex of the graph
 * starting form already computed coordinate systems at periodic points.
*/

#include "periodicPoints.h"

#ifndef DW_COORD_SYSTEMS
#define DW_COORD_SYSTEMS
/**
 * Implementation of the Algorithm 2 from the article.
 * 
 * Parameters:
 * GraphT G - a graph that encloses a map.
 * SetVertex checked[0] - (S in the article) - seed set, the vertexes that have
 *                     already assigned coordinates (from periodic points).
 * depth - maximal number of iterations. 
*/
template<class VectorT, class MatrixT, class GraphT>
void spreadCoordSystemsFromFixPoints(
    GraphT& G,
    typename GraphT::SetVertex checked[],
    int depth
  )
{
  typedef typename GraphT::VectorType VectorType;
  typedef typename GraphT::MatrixType MatrixType;
  typedef typename GraphT::Vertex Vertex;

  Vertex resolution = G.getResolution();
  Vertex denominator = GraphT::resolutionToDenominator(resolution);

  std::cout << "Algorithm 2 - spreading coordinates from periodic points.\n\n";
  int dim = resolution.dimension();
  // In the begining checked[0] contains already computed boxes.
  std::cout << "Boxes that have coordinates computed from periodic points: " << checked[0].size() << std::endl;

  // For each 'i' we spread coordinates from checked[i] to new boxes and we put them to checked[i+1]
  // The loop stops when checked[i+1].size()==0 
  for(int i=0;i<depth;++i)
  {
    typename GraphT::SetVertex::iterator b = checked[i].begin(), e = checked[i].end();
    int j = static_cast<int>(checked[i].size());
    #pragma omp parallel
    {
      // for each element in checked[i] proceed:
      #pragma omp for
      for(int q=0;q<j;++q)
      {
        // take a centre of a box
        typename GraphT::iterator current;
        #pragma omp critical
        {
          current = G.nodes().find(*b);
          ++b;
        }
        VectorT u(dim);

        // ---> here is some heurestic method for computing of coordinates - as described in the article.
        GraphT::centreOfBox(current->first,denominator,u);

        MatrixT prevCoord(dim,dim);
        intervalToDouble(current->second.m_coordSystem->begin(),current->second.m_coordSystem->end(),prevCoord.begin());
        // compute coord system
        MatrixT df(dim,dim), temp(dim,dim);
        int s;
        std::list<MatrixT> der;
        VectorT unstable;

        const int N=2;
        for(s=0;s<N;++s)
        {
          u = G.getMap()(u,temp);
          der.push_front(temp);
          prevCoord = temp*prevCoord;
          if(s==0)
            unstable = prevCoord.column(0);
        }

        capd::matrixAlgorithms::QR_decompose(prevCoord,df,temp);

        for(s=0;s<N-1;++s)
        {
          df = capd::matrixAlgorithms::gaussInverseMatrix(*(der.begin()))*df;
          der.pop_front();
        }
        df.column(0) = unstable;
        normalizeColumns(df);
        // ---< the end of heurestic method. Now we have df - new coordinate system.

        MatrixType M(dim,dim);
        std::copy(df.begin(),df.end(),M.begin());

        // We assign the coordinate system df to each vertex v
        // such that there is an edge (current,v)
        // and v does not have coordinate system, yet.
        typename GraphT::SetVertex::iterator k = current->second.m_edgesOut.begin();
        while(k!=current->second.m_edgesOut.end()) // foreach point mapped by (*b)
        {
          #pragma omp critical
          {
            typename GraphT::iterator it = G.nodes().find(*k);
            if(it->second.m_coordSystem==0) // if the node has no coord system, then proceed
            {
              it->second.m_coordSystem = new MatrixType(M);
              it->second.m_invCoordSystem = new MatrixType(capd::matrixAlgorithms::gaussInverseMatrix(M));
              checked[i+1].insert(*k); // and insert it as a new vertex 
            }
          }
          ++k;
        } // end while(k)
      } // end while(b)
    } // omp parallel
    std::cout << "coordinates spread to " << checked[i+1].size()  << " new vertexes " << std::endl;
    if(checked[i+1].size()==0) // the loop stops when there is no new vertexes
      break;
  } // end for (i)
}


// ----------------------- computeCoordSystems ------------------
/**
 * This routine computes coordinate systems in each vertex in the graph.
 * 
 * Parameters:
 * GraphT G - a graph that encloses a map.
 * perPointsFilename - file that contains numerical computed periodic points.
 * enclosureFilename - a file that contains an enclosure of the invariant set.
*/
template<const int depth, class VectorT, class MatrixT, class GraphT>
void computeCoordSystems(
    GraphT& G,
    const char perPointsFilename[],
    const char enclosureFilename[]
  )
{
  typename GraphT::SetVertex checked[100];
  // We call the routine for computing of coordinates at periodic points - in periodicPoints.h
  computeCoordSystemsAtFixPoints<depth,VectorT,MatrixT>(G,perPointsFilename,enclosureFilename,checked);

  // Then we call Algorithm 2 implemented above.
  spreadCoordSystemsFromFixPoints<VectorT,MatrixT>(G,checked, 99);

  // This is just for check if each vertex has assigned coordinate system.
  unsigned ch=0;
  for(int i=0;i<100;++i)
  {
    ch += checked[i].size();
  }
  std::cout << "coordinates set in " << ch << " boxes from " << G.nodes().size() << "\n\n";
}

#endif
