#include "capd/matrixAlgorithms/floatMatrixAlgorithms.hpp"

#ifndef DW_POS_DEF_MATRIX
#define DW_POS_DEF_MATRIX

/**
 * This is a rigorous routine. It verifies by means of the Sylvester criterion
 * if a symmetric matrix is positive definite.
 * The implementation is valid for matrices of dimension <=4, only.
*/
template<class MatrixT>
bool SylvesterCriterion(const MatrixT& M)
{
  int dim = M.numberOfRows();
  if(dim>4)
    return false;

  MatrixT Q(M.numberOfRows(),M.numberOfColumns());
  MatrixT R(M.numberOfRows(),M.numberOfColumns());
  typename MatrixT::ScalarType det;
  switch(dim)
  {
    case 4:
      // We have to compute determinant of the 4x4 matrix.
      // First we perform QR decomposition of M. 
      capd::matrixAlgorithms::QR_decompose(M,Q,R);
      // Then, the determinant of M is a product of the coefficients at the diagonal of R.
      det = R(1,1);
      for(int i=2;i<=4;++i)
      {
        det *= R(i,i);
      }
      // If det is not positive we return false.
      if(!(det>0))
        return false;

    case 3:
      // In 3x3 case we compute determinant by explicite formula.
      if(!(
           M(1,1)*M(2,2)*M(3,3) +
           M(1,2)*M(2,3)*M(3,1) +
           M(1,3)*M(2,1)*M(3,2) -
           M(3,1)*M(2,2)*M(1,3) -
           M(3,3)*M(2,1)*M(1,2) -
           M(1,1)*M(3,2)*M(2,3) > 0
          )
        )
        return false;

    case 2:
      // In 2x2 case we compute determinant by explicite formula.
      if(!(
           M(1,1)*M(2,2) - M(1,2)*M(2,1) > 0
          )
        )
        return false;

    default:
      if(!(M(1,1)>0))
        return false;
  }
  
  // All minors have positive detererminant, so the matrix is positive definite.
  return true;
}

/**
 * This is a rigorous routine.
 * It verifies positive definiteness of a matrix by J. Rohn method. 
 * It is enough to check if the corners of an interval matrix are positive definite. 
 * Then the interval matrix is positive definite too.
*/
template<class MatrixT>
bool checkPosDefMatrix(MatrixT M)
{
  MatrixT D(M.numberOfRows(),M.numberOfColumns());

  // Number of necessary corners to check is 2^(dim*(dim+1)/2)
  int size = (1 << (M.numberOfRows()*(M.numberOfRows()+1))/2);

  for(int k=0;k<size;++k)
  {
    // in the loops below we extract from the matrix its corner corresponding to the index 'k'
    int u=k;
    MatrixT Q(M.numberOfRows(),M.numberOfColumns());
    for(int p=1;p<=M.numberOfRows();++p)
    {
      for(int s=p;s<=M.numberOfColumns();++s)
      {
        if(u%2==0)
        {
          Q(p,s) = Q(s,p) = M(p,s).leftBound();
        }
        else
        {
          Q(p,s) = Q(s,p) = M(p,s).rightBound();
        }
        u/=2;
      }
    }
    
    // given a corner Q of the matrix, we verify positive definiteness of Q by Sylvester criterion.
    if(SylvesterCriterion(Q))
      continue;

    // if Sylvester does not work, try to use Gershgorin theorem.
    // First we rigorously diagonalize matrix, i.e. we compute a matrix with the same eigenvalues
    // but as close as possible to diagonal.
    if (!capd::matrixAlgorithms::symMatrixDiagonalize(Q,D,1e-12))
    {
      std::cout << "Cannot diagonalize a matrix\n";
      return false;
    }

    // Then verify if a_{ii} > sum_{i\neq j} |a_{ij}|
    for(int i=1;i<=M.numberOfRows();++i)
    {
      interval p=0;
      for(int j=1;j<=M.numberOfColumns();++j)
      {
        if(j!=i)
          p += capd::abs(D(i,j)); // sum coefficients in the row out of diagonal
      }
      if( !(D(i,i)>p) )
        return false;
    }
  }

  // All corners has been checked successfully, so the matrix is positive definite. 
  return true;
}


/**
 * This function computes A^T*Q*A - Q
 * If Q is diagonal this can be optimized.
 * The matrix Q is represented as vector.
*/
template<class MatrixT, class VectorT>
MatrixT AT_Q_A_minus_Q(const MatrixT& A, VectorT& Q)
{
  MatrixT result(A.numberOfRows(),A.numberOfColumns());
  result.clear();
  int i,j,k;
  for(i=1;i<=A.numberOfRows();++i)
  {
    for(k=1;k<=A.numberOfRows();++k)
    {
      result(i,i) += Q[k-1] * power(A(k,i),2);
    }
    result(i,i) -= Q[i-1];
  }

  for(i=1;i<=A.numberOfRows();++i)
  {
    for(j=i+1;j<=A.numberOfColumns();++j)
    {
      for(k=1;k<=A.numberOfRows();++k)
        result(i,j) += Q[k-1] * A(k,i)*A(k,j);
      result(j,i) = result(i,j);
    }
  }
  return result;
}

#endif
