#include <ctime>
#include <cstdlib>
#include <list>
#include <vector>
#include <omp.h>

// here we include all the implemented algorithms
#include "MapGraph.h"
#include "coordSystems.h"
#include "coneConditions.h"

// here is the class that represents the Poincare map of the Kuznetsov system.
#include "SmaleWilliamsMap.h"
typedef MapGraph<SmaleWilliamsMap,capd::vectalg::Vector<short int,4> > SmaleWilliamsMapGraph;

static const char swEnclosureFilename[] = "sw.dat";
static const char swCycleFilename[] = "swCycles";
static const char swPerPointsFilename[] = "swPerPoints";
static const char swCoordSysFilename[] = "swCoordSys";
static const int swDepth = 6;

short int resData[]={14,14,14,14};
double period = 6;
double A = 5;
double eps=0.5;
int defaultOrder=8;

// ----------------------- hyperbolicAttractor -------------------------
/**
 * This function verifies the existence of hyperbolic attractor
 * for the Poincare map in the Kuznetsov system.
 * 
 * The function just calls general algorithms for enclosing the attractor,
 * finding periodic points, computing of coordinate systems
 * and verification of the cone condition.
*/
void hyperbolicAttractor()
{
  SmaleWilliamsMapGraph::Vertex resolution(4,resData);
  interval omega=2*interval::pi();
  SmaleWilliamsMap sw(omega,period,A,eps,defaultOrder);
  SmaleWilliamsMapGraph G(sw,resolution);

  SWVector domain(4);
  domain[0] = domain[1] = domain[2] = domain[3] = interval(-2,2);

  // we take some initial point
  interval fixPoint[4]={
      interval(-0.22173209035238413102),
      interval(0.97492087818262221046),
      interval(0.018539814418960732734),
      interval(-0.03127925657964333927)
      };
  SWVector initPoint(4,fixPoint);

  // Call of the Algorithm 1 to enclose the attractor
  std::cout << "-------- Enclosing of attractor (Algorithm 1) ----------- \n";
  G.generateFirstCover(initPoint);
  G.save(swEnclosureFilename);
  std::cout << "done\n\n"; 
 
  std::cout << "----------------- Detecting cycles in the graph ------------- \n";
  // Call of the algorithm for detecting cycles in graph - in the file periodicPoints.h
  findCycles<swDepth>(G,swEnclosureFilename,swCycleFilename);
  std::cout << "done\n\n";

  std::cout << "----------------- Detecting fixed points ------------- \n";
  // Call of the algorithm for refining cycles to periodic points - in the file periodicPoints.h
  G.getMap().setOrder(20);
  refineCycles<swDepth,LDSWVector,LDSWMatrix>(G,swEnclosureFilename,swCycleFilename, swPerPointsFilename);
  std::cout << "done\n\n";

  std::cout << "----------------- Computing of coordinate systems ------------- \n";
  // In the following routine we compute cooordinate systems at the periodic points,
  // then we call Algorithm 2 for spreading coordinate system from periodic points.
  // Implementation in the files periodicPoints.h and coordSystems.h.
  computeCoordSystems<swDepth,LDSWVector,LDSWMatrix>(G,swPerPointsFilename, swEnclosureFilename);
  G.saveCoordSys(swCoordSysFilename);
  std::cout << "done\n\n";


  std::cout << "----------------- Verification of the cone condition (Algorithm 3) ------------- \n";
  // set parameters for rigorous solvers
  G.getMap().setOrder(10);
  sw.setStepControl(capd::dynsys::ILastTermsStepControl(1,1e-12));

  // This is quadratic form - one positive and three negative values
  SWVector Q;
  Q[0] = 1.;
  Q[1] = Q[2] = Q[3] = -1;

  // Call of the Algorithm 3 for verification of the cone condition - in coneCondition.h.
  verifyConeConditions(
      G, // graph
      Q, // quadratic form
      1, //subdivisionLevel, 1 means no subdivision
      "unverified", // filename to write unverified vertexes
      0, // index of the computer in the cluster
      1  // total number of computer in the cluster
    );
  std::cout << "done\n\n";
}

// ------------------- computeCover --------------------------------
/**
 * This is an auxiliary function. It computes boxes that cover the interval vector.
 * It is used to verify if the fixed point and periodic points belong to 
 * the attractor.
*/
void computeCover(SmaleWilliamsMap& s, SWVector x)
{
  SmaleWilliamsMapGraph::Vertex resolution(4,resData);
  SmaleWilliamsMapGraph G(s,resolution);

  SmaleWilliamsMapGraph::SetVertex cover;
  G.computeCover(cover,x);

  SmaleWilliamsMapGraph::SetVertex::iterator b = cover.begin();
  std::cout << "The interval Newton operator N is covered by the boxes: ";
  while(b!=cover.end())
  {
    std::cout << *b << " ";
    ++b;
  }
  std::cout << std::endl;
  // one can check using linux commands like grep and cat
  // that this vertexes belong to the cover of the attractor. Use:
  // cat sw.dat | grep eachElementOfTheAboveOutput  
}

// ------------------------ fixedPoint -----------------------------

/**
 * This function verifies the existence of a fixed in the attractor 
 * by means of the interval Newton method - see article for details.
*/
void fixedPoint()
{
  interval omega=2*interval::pi();
  SmaleWilliamsMap s(omega,period,A,eps,20/*order*/);
  s.setStepControl(capd::dynsys::ILastTermsStepControl(1));
  
  std::cout << "----------  The existence of a fixed point ---------- \n";
  // a candidate for fixed point.
  Real fp[] = {-0.2217320903523841312,0.97492087818262221051,0.018539814418960732759,-0.03127925657964333925};

  // The radius of a ball in which we verify the existence of fixed point.
  // Does not matter rigorous 1e-10 or not.
  Real diam = 1e-10;

  // first we compute image at centre
  SWVector centre;
  std::copy(fp,fp+4,centre.begin());

  // X is a ball or radius diam at centre
  SWVector X = centre;
  X += interval(-diam,diam);

  // compute P(centre)
  SWVector imCentre = s(centre);

  // compute DP(X)
  SWMatrix DP = s[X];

  // compute interval Newton operator
  DP -= SWMatrix::Identity(4);
  imCentre -= centre;
  SWVector N = - capd::matrixAlgorithms::gauss(DP,imCentre);
  if(subsetInterior(centre+N,X))
    std::cout << "The existence of fixed point verified\n";
  else
    std::cout << "The existence of fixed point not verified\n";

  // We have to verify if (centre+N) is a subset of attracting domain.
  // We compute boxes that cover the set (centre+N)
  computeCover(s,centre+N);
}

// ------------------------ periodTwoPoints -----------------------------
/**
 * This function verifies the existence of period two points in the attractor 
 * by means of the interval Newton method - see article for details.
*/
void periodTwoPoints()
{  
  interval omega=2*interval::pi();
  SmaleWilliamsMap s(omega,period,A,eps,20/*order*/);
  s.setStepControl(capd::dynsys::ILastTermsStepControl(1));

  int i,j;
  SWMatrix DP;
  capd::vectalg::Vector<interval,8> z, im, Z, NZ;
  capd::vectalg::Matrix<interval,8,8> DF;
  SWVector x,y;

  std::cout << "----------  The existence of period two points ---------- \n";

  // a candidate for period two orbit
  Real pt1[] = {1.0708168699357167158,0.093670228872794565237,0.015571609254819055212,0.024616361886390349154};
  Real pt2[] = {-0.22099700889399603573,-0.706209321733317795,-0.044189820507618084004,-0.006451355101601539937};
  
  // assign x and y
  std::copy(pt1,pt1+4,x.begin());
  std::copy(pt2,pt2+4,y.begin());
  
  // assign z=(x,y)
  std::copy(pt1,pt1+4,z.begin());
  std::copy(pt2,pt2+4,z.begin()+4);

  // Z is a ball centered as z of radius 1e-10, take rigorous if you wish
  Real diam = 1e-10;
  Z = z;
  Z += interval(-diam,diam);

  // here we compute F(x,y) = (P(x)-y,P(y)-x)
  SWVector imX = s(x) - y;
  SWVector imY = s(y) - x;
  // and embed the result into 8-dim vector 'im'
  std::copy(imX.begin(),imX.end(),im.begin());
  std::copy(imY.begin(),imY.end(),im.begin()+4);

  // x and y are unnecessary now. 
  // we make them balls centered at x and y to call
  // the routine for computing of derivative of Poincare map.
  x += interval(-diam,diam);
  y += interval(-diam,diam);
  // we compute derivative of P(x) and embed the result into a 8x8 matrix
  DP = s[x];
  for(i=1;i<=4;++i)
  {
    for(j=1;j<=4;++j)
    {
      DF(i,j) = DP(i,j); 
    }
    DF(i,i+4) = -1.;
  }
  
  // same at y, we compute derivative and embed it into 8x8 matrix
  DP = s[y];
  for(i=1;i<=4;++i)
  {
    for(j=1;j<=4;++j)
    {
      DF(i+4,j+4) = DP(i,j); 
    }
    DF(i+4,i) = -1.;
  }

  // eventually, we compute the interval Newton operator
  NZ = z - capd::matrixAlgorithms::gauss(DF,im);
  if(subsetInterior(NZ,Z))
    std::cout << "The existence of period two orbit verified\n";
  else
    std::cout << "The existence of period two orbit not verified\n";

  // We have to verify if pi_x(NZ) and pi_y(NZ) are subsets of attracting domain.
  // Take projections.
  std::copy(NZ.begin(),NZ.begin()+4,x.begin());
  std::copy(NZ.begin()+4,NZ.end(),y.begin());
  computeCover(s,x);
  computeCover(s,y);
}

// ---------------------------------- main -----------------------------

int main(int argc, char** argv)
{
  try{
    std::cout << "----------------- Proof of Theorem 1 ------------- \n";
    hyperbolicAttractor();

    std::cout << "----------------- Proof of Theorem 2 ------------- \n";
    fixedPoint();
    periodTwoPoints();
    
    std::cout << "Finished!\n";
  } catch(std::exception& e)
  {
    std::cout << "\nException caught!\n" << e.what() << std::endl;
  }
  return 0;
}

